"""Utility functions for working with interfaces."""

PARCIGNORE = [
    "unknown",
    "corpuscallosum",
    "Background+FreeSurfer_Defined_Medial_Wall",
    "???",
    "Unknown",
    "Medial_wall",
    "Medial wall",
    "medial_wall",
]
